/*
 * TrikebotController.java
 *
 * Created on April 12, 2002, 11:15 AM
 */

public class TrikebotController
{
	public Shell			root;

	public CalibrationConstants constants;
	int	panZero;
	int tiltZero;
	int wheelZero;

    public TrikebotState	state;
    public CMUcam			camera;
	public BrainStem		brainstem;

    public  TrackInfo   track;    
    private TrackThread trackThr;

	public	Macros		macros;
    
    private TrikebotCommand	command;
    private SerialPort		serialPort;
    private NetPort			controlPort;
    private NetPort			trackingPort;
    private boolean			serial;

    private static boolean DEBUG = false; // Print additional diagnostic information
    
	// Socket read timeout
    public final static int READ_TIMEOUT = 2000;
    
    /** Creates new TrikebotController */
    public TrikebotController(Shell r)
	{
		root		= r;
		constants	= new CalibrationConstants();
		FindZeros();
		state		= new TrikebotState();
        camera		= new CMUcam();
		brainstem	= new BrainStem();
        command		= new TrikebotCommand();
		track		= new TrackInfo();
        serialPort	= null;
        controlPort	= null;
        trackingPort= null;
		macros		= new Macros();
    }

    /* serial version */
    public boolean InitComm(int trikebotPort, int cameraPort)
	{
        int error;
        
        serial = true;
        serialPort = new SerialPort();
        
        if (DEBUG)
            System.err.println("Calling openSerial1(" + trikebotPort + ", 5)...");
        error = serialPort.openSerial1(trikebotPort, 5);
        System.err.println ("completed openSerial trikebot command, got " + error);
        if (error != 0)
		{
            System.err.println ("Error " + error + ": failed to connect to trikebot!");
            return false;
        }
        SetReadTimeout(READ_TIMEOUT);    
        System.err.println("Trikebot Serial Comm set up");

        if (!camera.InitCamera(cameraPort, 5))
		{
            System.err.println ("Failed to connect to CMUcam port!");
            return false;
        }
        System.err.println("InitComm completed!");

        return true;    
    }

    /* TCP/IP (wireless) version */
    public boolean InitComm(String ipaddr)
	{
        int error;
		boolean success = true;
        serial = false;
        controlPort = new NetPort();
        trackingPort = new NetPort();
        
        error = controlPort.OpenSocket(ipaddr, 5000);
        if (error != 0)
		{
            System.err.println("Error " + error + " in trikebot InitComm calling OpenSocket");
            success = false;
        }
		else
		{
			macros.SetPort(controlPort);
			if (DEBUG)
				System.err.println("Trikebot wireless socket set up!");
		}

        if (!camera.InitCamera(ipaddr, this))
		{
            System.err.println("Failed to init CMUcam!");
            success = false;
        }
        else if (DEBUG)
            System.err.println("CMUcam inited!");

        if (!brainstem.InitBrainStem(this))
		{
            System.err.println("Failed to init BrainStem!");
            success = false;
        }
        else if (DEBUG)
            System.err.println("BrainStem inited!");
        
        error = trackingPort.OpenSocket(ipaddr, 5002);
        if (error != 0)
		{
            System.err.println("Error " + error + " in trikebot InitComm calling OpenSocket for tracking");
            success = false;
        }
        else if (DEBUG)
            System.err.println("Tracking socket successfully opened!");

        SetReadTimeout(READ_TIMEOUT);
        
        return success;
    }
    
    public boolean InitRobot()
	{
        int i;

		if (!SendCommandFullResponse(command.ResetRobot()))
			return false;
		/*
        // call CheckCamera a limited number of times before giving up...
        for (i = 0; i < 5; i++)
            if (camera.CheckCamera())
                break; 
            else
			{
                System.err.println("CheckCamera failure..trying again");
                Time.Sleep(500);                
            }
        if (i == 5)
		{ 
          System.err.println ("Camera is not responding! Quitting...");
          return false;
        }
        System.err.println("CheckCamera worked. Now trying to reset camera");

        int iter = 0;
        while (!camera.ResetCamera())
		{
            System.err.println("ResetCamera failure. Retrying...");
            iter++;
            if (iter > 8)
			{
                System.err.println("PANIC: camera reset failure. Exiting");
                return false;
            }
        }
        System.err.println("Successfully reset camera.");

        FlushBuffer();
        SetParameters(true, 2, true, 0, true, 0, true, 0, true, 0);
        RefreshState();
        */
        return true;    
    }

    public boolean KillRobot()
	{
		controlPort.SendQuit();
		return true;
	}

    public boolean CloseComm()
	{
        int error;
		boolean success = true;

        if (serial && serialPort != null)
		{
            SleepServos(true, true);
            error = serialPort.closeSerial0();
            if (error != 0)
			{
                System.err.println("Error closing serial port!");
                success = false;
            }
        }
		else
		{
            if (controlPort != null)
			{
                SleepServos(true, true);
                error = controlPort.CloseSocket();
                if (error != 0)
				{
                    System.err.println("Error closing wireless socket!");
                    success = false;
                }
            }

            if (trackingPort != null)
			{
                error = trackingPort.CloseSocket();
                if (error != 0)
				{
                    System.err.println("Error closing tracking socket!");
                    success = false;
                }
            }
        }
        success = success && camera.ClosePorts();
		return success;
    }
    
    /* timeout, velocity, turn, camera pan, camera tilt */
    public boolean SetParameters(boolean tmset, int time, boolean vset, int vel, boolean tset, int turn,
								 boolean cpset, int pan, boolean ctset, int tilt)
	{
        boolean ans;
		int repeatCount = 0;
		do
		{
			ans = SendCommandFullResponse(command.SetParameters(tmset, time, vset, Speed2Pos(vel), tset, WheelAngle2Pos(turn),
													            cpset, PanAngle2Pos(pan), ctset, TiltAngle2Pos(tilt)));
		}
		while (++repeatCount < 3 && state.Status() == TrikebotState.BRAINSTEM_TIMEOUT);
		
        if (ans) 
		{	// Update the state parameters that are not available from the robot
			if (tset)
				state.SetWheelAngle(turn);
			if (cpset)
				state.SetPan(pan);
			if (ctset)
				state.SetTilt(tilt);
		}
		else
            FlushBuffer();
		
        return ans;
    }

	public boolean SetDriveTimeout(int timeout)
	{
        return SetParameters(true, timeout, false, 0, false, 0, false, 0, false, 0);
	}

    public boolean SetVel(int vel)
	{
        return SetParameters(false, 0, true, vel, false, 0, false, 0, false, 0);
    }

    public boolean KillMotor()
	{
        return SendCommandFullResponse(command.KillMotor());
    }
    
    public boolean TurnWheelTo(int turn)
	{
        return SetParameters(false, 0, false, 0, true, turn, false, 0, false, 0);
    }

    public boolean Drive(int vel, int turn)
	{
        return SetParameters(false, 0, true, vel, true, turn, false, 0, false, 0);
    }
    
    public boolean MoveHeadTo(int pan, int tilt)
	{
        return SetParameters(false, 0, false, 0, false, 0, true, pan, true, tilt);
	}

	public boolean PanHeadTo(int pan)
	{
		return SetParameters(false, 0, false, 0, false, 0, true, pan, false, 0);
	}

	public boolean TiltHeadTo(int tilt)
	{
		return SetParameters(false, 0, false, 0, false, 0, false, 0, true, tilt);
	}

	/** deprecated **/
	public boolean PanCameraTo(int pan)
	{
		return PanHeadTo(pan);
	}

	/** deprecated **/
	public boolean TiltCameraTo(int tilt)
	{
		return TiltHeadTo(tilt);
    }
    
    public boolean RefreshState()
	{
        return SendCommandFullResponse(command.SetParameters(false, 0, false, 0, false, 0, false, 0, false, 0));
    }

    public boolean SleepServos(boolean pan, boolean tilt)
	{
        return SetParameters(false, 0, false, 0, false, 0, pan, 0, tilt, 0);
    }

    /* like SetParameters, but expects direct servo values for pan, tilt, and wheel,
	 * and doesn't do angle to position conversion!
     * Should only be used for calibration...
     */
	public boolean DirectSetParameters(boolean tmset, int time, boolean vset, int vel, boolean tset, int turn,
									   boolean cpset, int pan, boolean ctset, int tilt)
	{
        boolean ans = SendCommandFullResponse(command.SetParameters(tmset, time, vset, vel, tset, turn, cpset, pan, ctset, tilt));
        if (!ans)
            FlushBuffer();
        return ans;
    }
    
    /* ******************** *
     * tracking on the iPaq *
     * ******************** */

    public void ActiveWindowTracking(boolean doPan, boolean doTilt)
	{
		if (doPan && doTilt)
			ActiveWindowTracking();
		else if (doPan)
			ActiveWindowTrackingNoTilt();
		else if (doTilt)
			ActiveWindowTrackingNoPan();
		else
			ActiveWindowTrackingNoMove();
	}
	
    public void ActiveWindowTracking()
	{
        DoActiveWindowTrackingWithCommand(command.ActiveWindowTracking());
    }

    public void ActiveWindowTrackingNoTilt()
	{
        DoActiveWindowTrackingWithCommand(command.ActiveWindowTrackingNoTilt());
    }

    public void ActiveWindowTrackingNoPan()
	{
        DoActiveWindowTrackingWithCommand(command.ActiveWindowTrackingNoPan());
    }

    public void ActiveWindowTrackingNoMove()
	{
        DoActiveWindowTrackingWithCommand(command.ActiveWindowTrackingNoMove());
    }

    public void ActiveColorTracking(int rmin, int rmax, int gmin, int gmax, int bmin, int bmax, boolean doPan, boolean doTilt)
	{
		if (doPan && doTilt)
			ActiveColorTracking(rmin, rmax, gmin, gmax, bmin, bmax);
		else if (doPan)
			ActiveColorTrackingNoTilt(rmin, rmax, gmin, gmax, bmin, bmax);
		else if (doTilt)
			ActiveColorTrackingNoPan(rmin, rmax, gmin, gmax, bmin, bmax);
		else
			ActiveColorTrackingNoMove(rmin, rmax, gmin, gmax, bmin, bmax);
	}

    public void ActiveColorTracking(int rmin, int rmax, int gmin, int gmax, int bmin, int bmax)
	{
        DoActiveTrackingWithCommand(command.ActiveColorTracking(rmin, rmax, gmin, gmax, bmin, bmax));
    }

    public void ActiveColorTrackingNoTilt(int rmin, int rmax, int gmin, int gmax, int bmin, int bmax)
	{
        DoActiveTrackingWithCommand(command.ActiveColorTrackingNoTilt(rmin, rmax, gmin, gmax, bmin, bmax));
    }

    public void ActiveColorTrackingNoPan(int rmin, int rmax, int gmin, int gmax, int bmin, int bmax)
	{
        DoActiveTrackingWithCommand(command.ActiveColorTrackingNoPan(rmin, rmax, gmin, gmax, bmin, bmax));
    }

    public void ActiveColorTrackingNoMove(int rmin, int rmax, int gmin, int gmax, int bmin, int bmax)
	{
        DoActiveTrackingWithCommand(command.ActiveColorTrackingNoMove(rmin, rmax, gmin, gmax, bmin, bmax));
    }

	// Does some preliminary work, then calls DoActiveTrackingWithCommand
	private void DoActiveWindowTrackingWithCommand(byte[] commandBytes)
	{
        System.err.println("Setting up the camera...");
        camera.CommandNoResponse("CR 18 44 19 33");  	// Auto White Balance On, Auto Gain On
        Time.Sleep(5000);								// Wait 5 seconds
        camera.CommandNoResponse("CR 18 40 19 32");  	// Auto White Balance Off, Auto Gain Off

		DoActiveTrackingWithCommand(commandBytes);		
	}

	// Ultimately used by all forms of ActiveTracking
	private void DoActiveTrackingWithCommand(byte[] commandBytes)
	{
		System.err.println("Initiating tracking");

        camera.CommandNoResponse("RM 1");	// Raw Mode On
        camera.CommandNoResponse("PM 0");	// Poll Mode Off
		SendCommandMiniResponse(commandBytes);

        trackThr = new TrackThread(this, trackingPort);
        trackThr.start();
	}

    public void StopActiveTracking()
	{
		SendCommandMiniResponse(command.StopActiveTracking());
        
        if (trackThr != null)
            trackThr.destroy();

        camera.CommandNoResponse("RM 0");	// Raw Mode Off
        camera.CommandNoResponse("PM 1");	// Poll Mode On
    }

    public void Safety(int reading)
	{
		SendCommandMiniResponse(command.Safety((byte) reading));
    }
	
	public int[] Sweep(int startAngle, int endAngle, int step)
	{
		int[] readings = new int[ (endAngle - startAngle) / step + 1];
		TurnWheelTo(startAngle);
		Time.Sleep(3500); // Takes a while to get there
		for (int th = startAngle, i = 0; th <= endAngle; th += step, i++)
		{
			RefreshState();
			readings[i] = state.Range();
			TurnWheelTo(th);
			Time.Sleep(500);
		}
		return readings;
	}
    
    /************************************
     * conversions between angle and    *
     * servo values                     *
     ************************************/

	public void FindZeros()
	{
		panZero = (0 - constants.PANMIN_ANGLE) * (constants.PANMAX - constants.PANMIN)
											   / (constants.PANMAX_ANGLE - constants.PANMIN_ANGLE) + constants.PANMIN;

        tiltZero = (0 - constants.TILTMIN_ANGLE) * (constants.TILTMAX - constants.TILTMIN)
												 / (constants.TILTMAX_ANGLE - constants.TILTMIN_ANGLE) + constants.TILTMIN;

        wheelZero = (0 - constants.TURNMIN_ANGLE) * (constants.TURNMAX - constants.TURNMIN)
			                                      / (constants.TURNMAX_ANGLE - constants.TURNMIN_ANGLE) + constants.TURNMIN;
	}
	
    public int PanPos2Angle(int pos)
	{
        return ((pos - panZero) * (constants.PANMAX_ANGLE - constants.PANMIN_ANGLE)) / (constants.PANMAX - constants.PANMIN);    
    }
  
    public int PanAngle2Pos(int angle)
	{
        int pos = (angle * (constants.PANMAX - constants.PANMIN)) / (constants.PANMAX_ANGLE - constants.PANMIN_ANGLE) + panZero;
        if (pos < constants.PANMIN)
			pos = constants.PANMIN;
        else if (pos > constants.PANMAX)
			pos = constants.PANMAX;
        return pos;    
    }

    public int TiltPos2Angle(int pos)
	{
        return ((pos - tiltZero) * (constants.TILTMAX_ANGLE - constants.TILTMIN_ANGLE)) / (constants.TILTMAX - constants.TILTMIN);
    }
  
    public int TiltAngle2Pos(int angle)
	{
        int pos = (angle * (constants.TILTMAX - constants.TILTMIN)) / (constants.TILTMAX_ANGLE - constants.TILTMIN_ANGLE) + tiltZero;
        if (pos < constants.TILTMIN)
			pos = constants.TILTMIN;
        else if (pos > constants.TILTMAX)
			pos = constants.TILTMAX;
        return pos;
    }

    public int WheelPos2Angle(int pos)
	{
        return ((pos - wheelZero) * (constants.TURNMAX_ANGLE - constants.TURNMIN_ANGLE))
		                          / (constants.TURNMAX - constants.TURNMIN);
    }

    public int WheelAngle2Pos(int angle)
	{
        int pos = (angle * (constants.TURNMAX - constants.TURNMIN)) / (constants.TURNMAX_ANGLE - constants.TURNMIN_ANGLE) + wheelZero;
        if (pos < constants.TURNMIN)
			pos = constants.TURNMIN;
        else if (pos > constants.TURNMAX)
			pos = constants.TURNMAX;
        return pos;
    }

    public int Pos2Speed(int pos)
	{
		if (pos > constants.WHEEL_FWD_MIN)
		{
			return ((pos - constants.WHEEL_FWD_MIN) * constants.WHEEL_FWD_SPEED) / (constants.WHEEL_FWD_MAX - constants.WHEEL_FWD_MIN);
		}
		else if (pos < constants.WHEEL_REV_MIN)
		{
			return ((pos - constants.WHEEL_REV_MIN) * constants.WHEEL_REV_SPEED) / (constants.WHEEL_REV_MAX - constants.WHEEL_REV_MIN);
		}
		else
		{
			return 0;
		}
	}

    public int Speed2Pos(int speed)
	{
		if (speed > 0)
		{
			int pos = ((speed * (constants.WHEEL_FWD_MAX - constants.WHEEL_FWD_MIN)) / constants.WHEEL_FWD_SPEED) + constants.WHEEL_FWD_MIN;
			if (pos > constants.WHEEL_FWD_MAX) // Cap at max speed
				return constants.WHEEL_FWD_MAX;
			else
				return pos;
		}
		else if (speed < 0)
		{
			int pos = ((speed * (constants.WHEEL_REV_MAX - constants.WHEEL_REV_MIN)) / constants.WHEEL_REV_SPEED) + constants.WHEEL_REV_MIN;
			if (pos < constants.WHEEL_REV_MAX) // Cap at max speed
				return constants.WHEEL_REV_MAX;
			else
				return pos;
		}
		else
			return (constants.WHEEL_FWD_MIN + constants.WHEEL_REV_MIN) / 2;
	}
    
    /***********************************
     * helper communications functions *
     ***********************************/
    public void FlushBuffer()
	{
        if (serial)
		{
            SetReadTimeout(10);
            while (serialPort.readByte1() != -1 );
			SetReadTimeout(READ_TIMEOUT);
        }
		else
		{
            controlPort.Flush();
        }
    }
    
    private void SetReadTimeout(int timeoutval)
	{
        if (serial)
            serialPort.setReadTimeout1(timeoutval);
        else
		{
            controlPort.SetReadTimeout(timeoutval);
            trackingPort.SetReadTimeout(timeoutval);
        }
    }

    private synchronized boolean SendCommandFullResponse(byte[] bytes)
	{
		FlushBuffer();
		return (SendCommand(bytes) && GetResponse(true));
	}

    private synchronized boolean SendCommandMiniResponse(byte[] bytes)
	{
		FlushBuffer();
		return (SendCommand(bytes) && GetResponse(false));
	}

    private synchronized String SendCommandStringResponse(byte[] bytes)
	{
		FlushBuffer();
		if (!SendCommand(bytes))
			return null;
		return GetStringResponse();
	}

	public String CMUcamDialogue(String s)
	{
		return SendCommandStringResponse(command.CMUcam(s));
	}

	public String BrainStemDialogue(byte[] b)
	{
		return SendCommandStringResponse(command.BrainStem(b));
	}

	public boolean SendDumpFramePacket()
	{
		return SendCommandMiniResponse(command.DumpFrame());
	}
	
    private boolean SendCommand(byte[] bytes)
	{
        int error;

        if (serial)
		{
            for (int i = 0; i < bytes.length; i++)
			{
                if (DEBUG)
                    System.err.println ("sending bytes: " + bytes[i]);
            
                error = serialPort.sendByte1(ByteUtil.Unsign(bytes[i]));
            
                if (error != 0)
				{
                    if (DEBUG)
                        System.err.println("Error in sendCommand: got " + error + " when sending the " + i + "th byte");
                    return false;
                }
            }
        }
		else
		{
           error = controlPort.SendBytes(bytes);
            if (error != 0)
			{
                System.err.println("Error sending trikebot command to iPaq!");
                return false;
            }
        }
		return true;
    }
    
    private boolean GetResponse(boolean returnsState)
	{
        byte[] bytes;

        if (serial)
		{
            bytes = new byte[state.FULL_PACKET_LENGTH];
            for (int i = 0; i < state.FULL_PACKET_LENGTH; i++)
			{
                int b = serialPort.readByte1();
                if (b == -1)
				{
                    if (DEBUG)
                        System.err.println("Error in GetResponse: got -1 on the " + i + "th byte");
                    return false;
                }
                
                bytes[i] = (byte) b;
            }
        }
		else
		{
            bytes = controlPort.ReadByteLine();
            if (bytes == null)
                return false;
        }

        boolean success;
		if (!returnsState || bytes.length == state.MINI_PACKET_LENGTH)
			success = state.ParseMiniPacket(bytes);
		else
			success = state.ParseFullPacket(bytes);
        if (!success)
		{
            if (DEBUG)
                System.err.println("Received packet; failed to parse!");
            return false;
        }
        return true;
    }

	private String GetStringResponse()
	{
        byte[] bytes;

        if (serial)
		{
            bytes = new byte[state.FULL_PACKET_LENGTH];
            for (int i = 0; i < state.FULL_PACKET_LENGTH; i++)
			{
                int b = serialPort.readByte1();
                if (b == -1)
				{
                    if (DEBUG)
                        System.err.println("Error in GetResponse: got -1 on the " + i + "th byte");
                    return null;
                }

                bytes[i] = (byte) b;
            }
        }
		else
		{
            bytes = controlPort.ReadByteLine();
            if (bytes == null)
                return null;
        }

        String s;
		if (bytes.length <= state.MINI_PACKET_LENGTH)
		{
			state.ParseMiniPacket(bytes);
            if (DEBUG)
                System.err.println("Received mini packet instead of string packet!");
			return null;
		}
		else
		{
			s = state.ParseStringPacket(bytes);
			if (s == null && DEBUG)
					System.err.println("Received packet; failed to parse!");
			return s;
		}
    }
}